#include <avr/interrupt.h>
#include <util/delay.h>
#include <stdint.h>
#include <stdbool.h>
#include <stdlib.h>

#include "7seg.h"
#include "adc.h"
#include "button.h"
#include "com.h"
#include "led.h"
#include "display.h"

#include "assets.h"

static uint16_t numbers[4][4];

static uint8_t cleanHoles(int8_t deltaX, int8_t deltaY);
static uint8_t moveCells(int8_t deltaX, int8_t deltaY);
static void randomFill();
static void showNumbers();
static void drawNumber(uint8_t row, uint8_t col, uint16_t value);
static void drawGrid();
static void drawArrowBorders();

static uint8_t getPressedButtons() {
	static uint8_t previousState[2] = { UNKNOWN, UNKNOWN };
	uint8_t currentState[2] = { sb_button_getState(BUTTON0), sb_button_getState(
			BUTTON1) };
	uint8_t r = 0;
	for (uint8_t i = 0; i < 2; i++) {
		r |= currentState[i] == PRESSED && previousState[i] == RELEASED ?
				(1 << i) : 0;
		previousState[i] = currentState[i];
	}
	return r;
}

static int8_t deltasX[4] = { -1, 0, 1, 0 };
static int8_t deltasY[4] = { 0, -1, 0, 1 };

void main() {
	sb_display_enable();

	sb_display_fillScreenFromFlash(splash);
	while (getPressedButtons() == 0);

	sb_display_fillScreen(NULL);

	for (uint8_t i = 0; i < 4; i++) {
		for (uint8_t j = 0; j < 4; j++) {
			numbers[i][j] = 0;
		}
	}
	randomFill();
	randomFill();
	drawGrid();
	showNumbers();
	drawArrowBorders();

	while (1) {
		uint8_t direction = 3 - sb_adc_read(POTI) * 4 / 1024;
		sb_display_drawBitmapFromFlash(4, 90, 2, 16, arrows[direction]);

		if (getPressedButtons() & 0x01) {
			uint8_t newNumber = moveCells(deltasX[direction],
					deltasY[direction]);

			showNumbers();
			if (newNumber) {
				randomFill();
				showNumbers();
			}
		}
	}
}

static void drawArrowBorders() {
	sb_display_draw(4, 87, 0xFF);
	sb_display_draw(4, 86, 0xFF);
	sb_display_draw(5, 87, 0xFF);
	sb_display_draw(5, 86, 0xFF);
	sb_display_draw(4, 109, 0xFF);
	sb_display_draw(4, 110, 0xFF);
	sb_display_draw(5, 109, 0xFF);
	sb_display_draw(5, 110, 0xFF);
	sb_display_draw(3, 86, 0xF0);
	sb_display_draw(3, 87, 0xF0);
	sb_display_draw(3, 109, 0xF0);
	sb_display_draw(3, 110, 0xF0);
	sb_display_draw(6, 86, 0xF);
	sb_display_draw(6, 87, 0xF);
	sb_display_draw(6, 109, 0xF);
	sb_display_draw(6, 110, 0xF);
	for (uint8_t i = 0; i < 21; i++) {
		sb_display_draw(3, 88 + i, 0x30);
	}
	for (uint8_t i = 0; i < 21; i++) {
		sb_display_draw(6, 88 + i, 0xC);
	}
}

static uint8_t cleanHoles(int8_t deltaX, int8_t deltaY) {
	uint8_t change = 1;
	uint8_t changeHappened = 0;

	while (change != 0) {
		change = 0;
		for (uint8_t i = 0; i < 4; i++) {
			for (uint8_t j = 0; j < 4; j++) {
				if (j + deltaX < 0 || j + deltaX >= 4 || i + deltaY < 0
						|| i + deltaY >= 4) {
					continue;
				}
				if (numbers[i + deltaY][j + deltaX] == 0
						&& numbers[i][j] != 0) {
					numbers[i + deltaY][j + deltaX] = numbers[i][j];
					numbers[i][j] = 0;
					change = 1;
					changeHappened = 1;
				}
			}
		}
	}

	return changeHappened;
}

static uint8_t moveCells(int8_t deltaX, int8_t deltaY) {
	uint8_t changeHappened = cleanHoles(deltaX, deltaY);

	for (uint8_t i = 0; i < 4; i++) {
		for (uint8_t j = 0; j < 4; j++) {
			if (j + deltaX < 0 || j + deltaX >= 4 || i + deltaY < 0
					|| i + deltaY >= 4) {
				continue;
			}
			if (numbers[i + deltaY][j + deltaX] == numbers[i][j]) {
				numbers[i + deltaY][j + deltaX] += numbers[i][j];
				numbers[i][j] = 0;
				changeHappened = 1;
			}
		}
	}

	return changeHappened | cleanHoles(deltaX, deltaY);
}

static void randomFill() {
	uint8_t filled = 0;
	while (filled == 0) {
		uint8_t r = (uint8_t) rand() & 0x0F;
		uint8_t r2 = (uint8_t) rand() & 0x0F;

		if (numbers[r / 4][r % 4] == 0) {
			if (r2 == 15) {
				numbers[r / 4][r % 4] = 4;
			} else {
				numbers[r / 4][r % 4] = 2;
			}
			filled = 1;
		}
	}
}

static void showNumbers() {
	static uint16_t values[4][4];
	for (uint8_t i = 0; i < 4; i++) {
		for (uint8_t j = 0; j < 4; j++) {
			if (values[i][j] != numbers[i][j])
				drawNumber(i, j, numbers[i][j]);
			values[i][j] = numbers[i][j];
		}
	}
}

static void drawNumber(uint8_t row, uint8_t col, uint16_t value) {
	uint8_t number[16];
	for (uint8_t i = 0; i < 16; i++) {
		number[i] = 0;
	}
	switch (value) {

	case 2048:
		for (uint8_t i = 0; i < 4; i++) {
			number[i] = font0[2][i];
		}
		for (uint8_t i = 0; i < 4; i++) {
			number[i + 4] = font0[0][i];
		}
		for (uint8_t i = 0; i < 4; i++) {
			number[i + 8] = font0[4][i];
		}
		for (uint8_t i = 0; i < 4; i++) {
			number[i + 12] = font0[8][i];
		}
		break;
	case 1024:
		for (uint8_t i = 0; i < 4; i++) {
			number[i] = font0[1][i];
		}
		for (uint8_t i = 0; i < 4; i++) {
			number[i + 4] = font0[0][i];
		}
		for (uint8_t i = 0; i < 4; i++) {
			number[i + 8] = font0[2][i];
		}
		for (uint8_t i = 0; i < 4; i++) {
			number[i + 12] = font0[4][i];
		}
		break;
	case 512:
		for (uint8_t i = 0; i < 5; i++) {
			number[i + 1] = font1[5][i] << 1;
		}
		for (uint8_t i = 0; i < 5; i++) {
			number[i + 5 + 1] = font1[1][i] << 1;
		}
		for (uint8_t i = 0; i < 5; i++) {
			number[i + 10 + 1] = font1[2][i] << 1;
		}
		break;
	case 256:
		for (uint8_t i = 0; i < 5; i++) {
			number[i + 1] = font1[2][i] << 1;
		}
		for (uint8_t i = 0; i < 5; i++) {
			number[i + 5 + 1] = font1[5][i] << 1;
		}
		for (uint8_t i = 0; i < 5; i++) {
			number[i + 10 + 1] = font1[6][i] << 1;
		}
		break;
	case 128:
		for (uint8_t i = 0; i < 5; i++) {
			number[i + 1] = font1[1][i] << 1;
		}
		for (uint8_t i = 0; i < 5; i++) {
			number[i + 5 + 1] = font1[2][i] << 1;
		}
		for (uint8_t i = 0; i < 5; i++) {
			number[i + 10 + 1] = font1[8][i] << 1;
		}
		break;
	case 64:
		for (uint8_t i = 0; i < 6; i++) {
			number[i + 2] = font2[6][i];
		}
		for (uint8_t i = 0; i < 6; i++) {
			number[i + 6 + 2] = font2[4][i];
		}
		break;
	case 32:
		for (uint8_t i = 0; i < 6; i++) {
			number[i + 2] = font2[3][i];
		}
		for (uint8_t i = 0; i < 6; i++) {
			number[i + 6 + 2] = font2[2][i];
		}
		break;
	case 16:
		for (uint8_t i = 0; i < 6; i++) {
			number[i + 2] = font2[1][i];
		}
		for (uint8_t i = 0; i < 6; i++) {
			number[i + 6 + 2] = font2[6][i];
		}
		break;
	case 8:
		for (uint8_t i = 0; i < 6; i++) {
			number[i + 5] = font2[8][i];
		}
		break;
	case 4:
		for (uint8_t i = 0; i < 6; i++) {
			number[i + 5] = font2[4][i];
		}
		break;
	case 2:
		for (uint8_t i = 0; i < 6; i++) {
			number[i + 5] = font2[2][i];
		}
		break;
	case 1:
		for (uint8_t i = 0; i < 6; i++) {
			number[i + 5] = font2[1][i];
		}
		break;
	}

	for (uint8_t i = 0; i < 16; i++) {
		sb_display_draw(row * 2, col * 19 + 1 + i, (number[i] << 4) | 0x01);
		sb_display_draw(row * 2 + 1, col * 19 + 1 + i, number[i] >> 4 | 0x80);
	}
}

static void drawGrid() {
	for (uint8_t row = 0; row < 4; row++) {
		for (uint8_t col = 0; col < 4; col++) {
			sb_display_draw(row * 2, col * 19 + 0, 0xFF);
			sb_display_draw(row * 2 + 1, col * 19 + 0, 0xFF);
			sb_display_draw(row * 2, col * 19 + 18, 0xFF);
			sb_display_draw(row * 2 + 1, col * 19 + 18, 0xFF);
			for (uint8_t i = 1; i < 18; i++) {
				sb_display_draw(row * 2, col * 19 + i, 0x01);
				sb_display_draw(row * 2 + 1, col * 19 + i, 0x80);
			}
		}
	}
}

